import asyncio
from collections.abc import Iterable
import os
import subprocess
import tempfile
from typing import Any, Union

import feather
import pandas as pd
from tqdm.auto import tqdm
import typer

from solsim.system import BaseSystem, BaseSolanaSystem
from solsim.type import StateType


class Simulation:

    INDEX_COLS = ["step", "run"]

    def __init__(self, system: Union[BaseSystem, BaseSolanaSystem], watchlist: Iterable[str]) -> None:
        self._system = system
        self._watchlist = set(watchlist)

    @property
    def cli(self) -> typer.Typer:
        app = typer.Typer()

        @app.command()  # type: ignore
        def run(runs: int = 1, steps_per_run: int = 1, viz_results: bool = False) -> pd.DataFrame:
            return self.run(runs, steps_per_run, viz_results)

        @app.callback()  # type: ignore
        def callback() -> None:
            pass

        return app

    def run(self, runs: int = 1, steps_per_run: int = 3, visualize_results: bool = False) -> pd.DataFrame:
        """Run your simulation.

        Args:
            runs: The number of times to run your simulation.
            visualize_results: Optionally build and start a Streamlit app to explore simulation results.

        Returns:
            results: A pandas DataFrame containing your simulation results.
        """
        results = asyncio.run(self._run(runs, steps_per_run))
        if visualize_results:
            try:
                with tempfile.TemporaryDirectory() as tmpdir:
                    app = self._start_results_app(results, tmpdir)
                    app.wait()
            except KeyboardInterrupt:
                pass
        return results

    @staticmethod
    def _start_results_app(results: pd.DataFrame, tmpdir: str) -> subprocess.Popen[Any]:
        results_path = os.path.join(tmpdir, "results.feather")
        feather.write_dataframe(results, results_path)
        env = {**os.environ, "SOLSIM_RESULTS_PATH": results_path}
        return subprocess.Popen(["streamlit", "run", "visualize.py"], cwd=os.path.dirname(__file__), env=env)

    async def _run(self, runs: int, steps_per_run: int) -> pd.DataFrame:
        results: list[StateType] = []
        try:
            for run in range(runs):
                try:
                    state: StateType = {}
                    history: list[StateType] = []
                    self._system.setup()
                    for step in tqdm(range(steps_per_run), desc=f"🟢 run: {run} | step"):
                        if self._system.uses_solana:
                            updates = await self._system.initial_step() if step == 0 else await self._system.step(state, history)  # type: ignore  # noqa: E501
                        else:
                            updates = self._system.initial_step() if step == 0 else self._system.step(state, history)
                        state = {**state, **updates, "run": run, "step": step}
                        history.append(state)
                        results.append(self._filter_state(state))
                finally:
                    self._system.teardown()
        finally:
            await self._system.cleanup() if self._system.uses_solana else self._system.cleanup()

        results = pd.DataFrame(results)
        return self._reorder_results_columns(results)

    def _filter_state(self, state: StateType) -> StateType:
        for qty in self._watchlist:
            if qty not in state:
                raise Exception(f"{qty} not found in state: {state}")
        return {qty: state[qty] for qty in self._watchlist | set(self.INDEX_COLS)}

    def _reorder_results_columns(self, results: pd.DataFrame) -> pd.DataFrame:
        cols = self.INDEX_COLS + sorted([col for col in results.columns if col not in self.INDEX_COLS])
        return results[cols]



from spl.token.constants import TOKEN_PROGRAM_ID, ASSOCIATED_TOKEN_PROGRAM_ID  # noqa: F401
from solana.system_program import SYS_PROGRAM_ID  # noqa: F401
from solana.sysvar import SYSVAR_RENT_PUBKEY  # noqa: F401



from abc import ABC, abstractmethod
import time
import psutil
import signal
import subprocess
from subprocess import DEVNULL
import tempfile
from typing import Awaitable, Optional, List, Any

from anchorpy import close_workspace, create_workspace
from psutil import Process
from solana.publickey import PublicKey
from solana.rpc.api import Client
from solana.rpc import commitment

from solsim.type import StateType


class BaseMixin:
    @property
    def uses_solana(self) -> bool:
        return isinstance(self, BaseSolanaSystem)

    def setup(self) -> Any:
        pass

    def teardown(self) -> Any:
        pass

    def cleanup(self) -> Any:
        pass


class BaseSystem(ABC, BaseMixin):
    @abstractmethod
    def initial_step(self) -> StateType:
        """Return initial system state.

        Returns:
            The initial state of the system.
        """
        raise NotImplementedError

    @abstractmethod
    def step(self, state: StateType, history: List[StateType]) -> StateType:
        """Return an arbitrary state of the system.

        This method is `async` because it (presumably) will make RPC calls
        to the Solana blockchain, which are `async`.

        Args:
            state: The previous system state.
            history: The full history of system states.

        Returns:
            The initial state of the system.
        """
        raise NotImplementedError


class BaseSolanaSystem(ABC, BaseMixin):

    SOLANA_CLUSTER_URI = "http://127.0.0.1:8899"

    def __init__(
        self, workspace_dir: str, client: Optional[Client] = None, localnet_process: Optional[Process] = None
    ) -> None:
        self._workspace_dir = workspace_dir
        self._localnet_process = localnet_process
        self._localnet_initialized = False
        self.setup()
        self.client = client or Client(self.SOLANA_CLUSTER_URI)
        self.workspace = create_workspace(self._workspace_dir)

    def setup(self) -> None:
        if not self._localnet_initialized:
            if not self._localnet_process:
                self._logfile = tempfile.NamedTemporaryFile()
                self._localnet = self._start_localnet()
                print("👆 Starting Solana localnet cluster (~5s) ...")
                while not self._localnet_ready:
                    time.sleep(1)
            else:
                self._localnet = self._localnet_process
            self._localnet_initialized = True

    def teardown(self) -> None:
        if self._localnet_initialized:
            print("👇 Terminating Solana localnet cluster ...")
            self._terminate_localnet()
        self._localnet_initialized = False

    async def cleanup(self) -> None:
        await close_workspace(self.workspace)

    @abstractmethod
    async def initial_step(self) -> Awaitable[StateType]:
        """Return initial system state.

        This method is `async` because it (presumably) will make RPC calls
        to the Solana blockchain, which are `async`.

        Returns:
            The initial state of the system.
        """
        raise NotImplementedError

    @abstractmethod
    async def step(self, state: StateType, history: List[StateType]) -> Awaitable[StateType]:
        """Return an arbitrary state of the system.

        This method is `async` because it (presumably) will make RPC calls
        to the Solana blockchain, which are `async`.

        Args:
            state: The previous system state.
            history: The full history of system states.

        Returns:
            The initial state of the system.
        """
        raise NotImplementedError

    def _start_localnet(self) -> subprocess.Popen[Any]:
        for proc in psutil.process_iter():
            if proc.name() == "solana-test-validator":
                self._terminate_processes([proc])
        return subprocess.Popen(["anchor", "localnet"], cwd=self._workspace_dir, stdout=self._logfile, stderr=DEVNULL)

    @property
    def _localnet_ready(self) -> bool:
        lastline = subprocess.check_output(["tail", "-n", "1", self._logfile.name]).decode("utf-8").strip()
        return "| Processed Slot: " in lastline

    def get_token_account_balance(
        self, pubkey: PublicKey, commitment: Optional[commitment.Commitment] = commitment.Confirmed
    ) -> float:
        """Get account token balance.

        Args:
            pubkey: The public key of the account in question.

        Returns:
            The token balance of the account.
        """
        return float(self.client.get_token_account_balance(pubkey, commitment)["result"]["value"]["uiAmount"])

    def _terminate_processes(self, kill_list: list[Process], timeout: int = 10) -> None:
        # Attempt graceful termination first.
        for p in reversed(kill_list):
            self._signal_process(p, signal.SIGTERM)
        _, alive = psutil.wait_procs(kill_list, timeout=timeout)

        # Forcefully terminate procs still running.
        for p in alive:
            self._signal_process(p, signal.SIGKILL)
        _, alive = psutil.wait_procs(kill_list, timeout=timeout)

        if alive:
            raise Exception(f"could not terminate process {alive}")

    def _terminate_localnet(self) -> None:
        """
        Borrowed from https://github.com/pytest-dev/pytest-xprocess/blob/6dac644e7b6b17d9b970f6e9e2bf2ade539841dc/xprocess/xprocess.py#L35.  # noqa E501
        """
        parent = psutil.Process(self._localnet.pid)
        try:
            kill_list = [parent]
            kill_list += parent.children(recursive=True)
            self._terminate_processes(kill_list)
        except (psutil.Error, ValueError) as err:
            raise Exception(f"Error while terminating process {err}")
        else:
            return

    def _signal_process(self, p: Process, sig: signal.Signals) -> None:
        """
        Borrowed from: https://github.com/pytest-dev/pytest-xprocess/blob/6dac644e7b6b17d9b970f6e9e2bf2ade539841dc/xprocess/xprocess.py#L29.  # noqa E501
        """
        try:
            p.send_signal(sig)
        except psutil.NoSuchProcess:
            pass



# type: ignore

import os
import altair as alt
import feather
import streamlit as st

from solsim.simulation import Simulation


results = feather.read_dataframe(os.environ["SOLSIM_RESULTS_PATH"])
idx_cols = Simulation.INDEX_COLS
watched_vars = [col for col in results.columns if col not in idx_cols]

# Sidebar

st.sidebar.image(os.path.join(os.path.dirname(__file__), "../../img/logo.png"))

st.sidebar.markdown("# 👉 Select quantities")
quantities = st.sidebar.multiselect("State variables or KPIs to explore", watched_vars)
st.sidebar.markdown(
    f"""
    # 🧮 Metadata

    - **Runs: {results["run"].max() + 1}**
    - **Steps per run: {results["step"].max() + 1}**
    - **Watched quantities: {len(watched_vars)}**
"""
)

st.markdown("# 🔬 Results Explorer")

if quantities:

    st.markdown("## 📈 Graph")

    results = results[quantities + idx_cols]
    charts = []

    for run, group in results.groupby("run"):
        chart = (
            alt.Chart(group.melt(idx_cols))
            .mark_line()
            .encode(
                x=alt.X("step", axis=alt.Axis(tickMinStep=1), title="Timestep"),
                y=alt.Y(
                    "value",
                    title="Value",
                ),
                color="variable",
            )
            .properties(title="Simulation Quantities over Time")
        )
        charts.append(chart)

    st.altair_chart(alt.layer(*charts), use_container_width=True)

    st.markdown("## 📝 Table")

    st.table(results)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['solsim']

package_data = \
{'': ['*']}

install_requires = \
['anchorpy>=0.6.4,<0.7.0',
 'feather-format>=0.4.1,<0.5.0',
 'numpy>=1.22.1,<2.0.0',
 'pandas>=1.4.0,<2.0.0',
 'streamlit>=1.5.0,<2.0.0',
 'tqdm>=4.62.3,<5.0.0']

setup_kwargs = {
    'name': 'solsim',
    'version': '0.1.7',
    'description': 'The Solana complex systems simulator.',
    'long_description': '# solsim\n\n<div align="center">\n    <img src="https://raw.githubusercontent.com/cavaunpeu/solsim/main/img/logo.png" width="70%" height="70%">\n</div>\n\n---\n\n<!-- Badges -->\n\n[![Discord Chat](https://img.shields.io/discord/889577356681945098?color=blueviolet)](https://discord.gg/sxy4zxBckh)\n[![Downloads](https://pepy.tech/badge/solsim)](https://pepy.tech/project/solsim)\n\n<!-- Tests and Lint Badges -->\n\n<div>\n<img alt="Tests" src="https://github.com/cavaunpeu/solsim/actions/workflows/tests.yml/badge.svg">\n<img alt="Lint" src="https://github.com/cavaunpeu/solsim/actions/workflows/lint.yml/badge.svg">\n</div>\n\n---\n\n## Introduction\n\nsolsim is the Solana complex systems simulator. It simulates behavior of dynamical systems—DeFi protocols, DAO governance, cryptocurrencies, and more—built on the [Solana](https://solana.com/) blockchain.\n\n## Philosophy\n\nDefine your system how you see fit.\n\nsolsim will simulate its behavior and collect its results in a structured, straightforward manner.\n\n## Usage\n\n1. Implement `initial_step` and `step` methods.\n2. From each, return the current state, i.e. a dictionary mapping variables to current values.\n3. Specify the variables you\'d like to "watch."\n4. Instantiate a Simulation, call `.run()`.\n5. Receive a [pandas](https://pandas.pydata.org/) DataFrame containing values of "watched" variables at each step in time.\n\n### With Solana\n\n```python\nfrom anchorpy import Context\nfrom solana.keypair import Keypair\nfrom solsim.simulation import Simulation\n\nclass SomeSolanaSystem(BaseSolanaSystem):\n    def __init__(self):\n        super().__init__("path/to/workspace")\n        self.account = Keypair()\n        self.pubkey = self.account.public_key\n        self.program = self.workspace["my_anchor_program"]  # solsim gives a Anchor program workspace (self.workspace).\n\n    async def initial_step(self):\n        self.program.rpc["initialize"]()  # Make RPC calls to your Anchor program.\n        await self.client.request_airdrop(self.pubkey, 10)  # solsim gives you a Solana API client (self.client).\n        return {"balance": await self.client.get_balance(self.pubkey)}\n\n    async def step(self, state, history):\n        self.program.rpc["submit_uniswap_trade"](\n            ctx=Context(accounts={"account": self.pubkey}, signers=[self.account])\n        )\n        return {"balance": await self.client.get_balance(self.account)}\n\n\nsimulation = Simulation(system=SomeSolanaSystem(), watchlist=("balance"))\nresults = simulation.run(steps_per_run=5)  # Returns pandas DataFrame of results.\n```\n\n### Without Solana\n\n```python\nclass SomeSystem(BaseSystem):\n    def __init__(self, population):\n        self.pop = population\n\n    def initial_step(self):\n        return {"population": self.pop}\n\n    def step(self, state, history):\n        return {"population": state["population"] * 1.1}\n\n\nsimulation = Simulation(system=SomeSystem(), watchlist=("population"))\nresults = simulation.run(steps_per_run=5)\n```\n\n## CLI\n\nSimulations can also be run via CLI. Instead of calling `simulation.run()`, simply:\n\n1. Call `simulation.cli()`\n2. Run your simulation as e.g. `python path/to/file.py run --num-runs 3`\n\n## Results Explorer\n\nsolsim gives you a streamlit app to explore results, e.g.\n\n<div>\n    <img src="https://raw.githubusercontent.com/cavaunpeu/solsim/main/img/results_explorer_app.png">\n</div>\n\nTo automatically start this app following simulation, invoke one of the following:\n\n- `simulation.run(visualize_results=True)`\n- `--viz-results` flag in the CLI runner, e.g. `python path/to/file.py run --viz-results`\n\n## Installation\n\nFirst, install [Anchor](https://project-serum.github.io/anchor/getting-started/installation.html#install-rust).\n\n### Library\n\n```sh\npip install solsim\n```\n\n### Development\n\nInstall [poetry](https://python-poetry.org/). Then,\n\n```sh\ngit clone --recurse-submodules https://github.com/cavaunpeu/solsim.git\ncd solsim\npoetry install\npoetry shell\n```\n\n## Detailed Usage\n\n### With Solana\n\nFirst, write your Solana program. solsim prefers you do this in [Anchor](https://project-serum.github.io/anchor/getting-started/introduction.html). Then,\n\n1. Write a system class that inherits from `BaseSolanaSystem`.\n2. Call `super().__init__("path/to/program")` in its `__init__`.\n3. Implement `initial_step` and `step` methods. (Since you\'ll interact with Solana asynchronously, these methods should be `async`.)\n\nIn `2.`, solsim exposes the following attributes to your system instance:\n\n- `self.workspace`: IDL clients for the Solana programs that comprise your system (via [anchorpy](https://github.com/kevinheavey/anchorpy)).\n\nFor example, these clients let you interact with your respective programs\' RPC endpoints.\n\n- `self.client`: a general Solana client (via [solana-py](https://github.com/michaelhly/solana-py)).\n\nThis client lets you interact with Solana\'s RPC endpoints. Documentation [here](https://michaelhly.github.io/solana-py/api.html#).\n\nFinally,\n\n1. Define a `watchlist`: variables (returned in `initial_step` and `step`) you\'d like to "watch."\n2. Instantiate and run your simulation, e.g. `Simulation(MySystem(), watchlist).run(steps_per_run=10)`.\n\n### Without Solana\n\n1. Write a system class that inherits from `BaseSystem`.\n2. Implement `initial_step` and `step` methods.\n3. Define a `watchlist`.\n4. Instantiate and run your simulation.\n\n## Examples\n\n### Drunken Escrow\n\nAgents are randomly paired to exchange random amounts of `foo_coin` and `bar_coin` via an Anchor escrow contract in each timestep.\n\n- Run: `python -m examples.drunken_escrow`.\n- Code: [here](https://github.com/cavaunpeu/solsim/tree/main/examples/drunken_escrow).\n- Expected output (numbers may vary):\n\n```\n(.venv) ➜  solsim git:(main) $ python -m examples.drunken_escrow\nWaiting for Solana localnet cluster to start (~10s) ...\nSteps completed: 100%|████████████████████████████████████████████████████████████████████████████████████| 4/4 [00:27<00:00,  6.82s/it]\n   step  mean_balance_spread  mean_swap_amount  num_swaps\n0    -1            40.000000         30.666667          3\n1     0            58.000000         12.000000          3\n2     1            60.666667          4.000000          3\n3     2            83.333333         21.500000          2\n```\n\n### Lotka-Volterra\n\nThe Lotka-Volterra model is a classic dynamical system in the field of ecology that tracks the evolution of interdependent predator and prey populations.\n\n- Run: `python -m examples.lotka_volterra`.\n- Code: [here](https://github.com/cavaunpeu/solsim/tree/main/examples/lotka_volterra).\n- Expected output:\n\n```\n(.venv) ➜  solsim git:(main) ✗ python -m examples.lotka_volterra\nSteps completed: 100%|█████████████████████████████████████████████████████████████████████████████████| 4/4 [00:00<00:00, 28581.29it/s]\n   step  food_supply  population_size\n0    -1     1000.000            50.00\n1     0      995.000            60.00\n2     1      989.000            69.95\n3     2      982.005            79.84\n```\n\nThis implementation inspired by [cadCAD Edu](https://www.cadcad.education/).\n\n## Inspiration\n\nsolsim humbly builds on the shoulders of the giants that are [cadCAD](https://github.com/cadCAD-org/cadCAD) and [tokenspice](https://github.com/tokenspice/tokenspice), among others.',
    'author': 'Will Wolf',
    'author_email': 'williamabrwolf@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cavaunpeu/solsim',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
